#include <iostream>
#include <boost/thread/thread.hpp>
#include <pcl/point_cloud.h>
#include <pcl/io/pcd_io.h>
#include <pcl/point_types.h>
#include <pcl/features/normal_3d.h>
#include <pcl/features/fpfh.h>
#include <pcl/visualization/pcl_visualizer.h>

int main (int argc, char** argv)
{

  if(argc < 3){
    std::cout <<"Usage: <scale factor in cm> <name.pcd> [show normals?]\n";
    return(0);
  }
  
  double scaleFactor = atof(argv[1]) * .01;

  //Load the point cloud
  pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud (new pcl::PointCloud<pcl::PointXYZRGB>);

  if (pcl::io::loadPCDFile<pcl::PointXYZRGB> (argv[2], *cloud) == -1) //* load the file
  {
    printf ("Couldn't input file \n");
    return (-1);
  }
  std::cout << "Loaded "<< cloud->width * cloud->height<< " data points\n";
            
  
  // Create the normal estimation class, and pass the input dataset to it
  pcl::NormalEstimation<pcl::PointXYZRGB, pcl::Normal> ne;
  ne.setInputCloud (cloud);

  // Create an empty kdtree representation, and pass it to the normal estimation object.
  // Its content will be filled inside the object, based on the given input dataset (as no other search surface is given).
  pcl::search::KdTree<pcl::PointXYZRGB>::Ptr tree (new pcl::search::KdTree<pcl::PointXYZRGB> ());
  ne.setSearchMethod (tree);

  // Output datasets
  pcl::PointCloud<pcl::Normal>::Ptr normals (new pcl::PointCloud<pcl::Normal>);

  // Use all neighbors in a sphere of radius scaleFactor cm
  ne.setRadiusSearch (scaleFactor);

  // Compute the normals
  ne.compute (*normals);
  
  // Create the FPFH estimation class, and pass the input dataset+normals to it
  pcl::FPFHEstimation<pcl::PointXYZRGB, pcl::Normal, pcl::FPFHSignature33> fpfh;
  fpfh.setInputCloud (cloud);
  fpfh.setInputNormals (normals);
  fpfh.setSearchMethod (tree);

  // Output datasets
  pcl::PointCloud<pcl::FPFHSignature33>::Ptr fpfhs (new pcl::PointCloud<pcl::FPFHSignature33> ());

  // Use all neighbors in a sphere
  // IMPORTANT: the radius used here has to be larger than the radius used to estimate the surface normals!!!
  fpfh.setRadiusSearch (scaleFactor+.02);

  // Compute the features
  fpfh.compute (*fpfhs);
  
  //write the features to a file
  std::ofstream outfile;
  outfile.open(strcat(argv[2],".features"));
  if(!outfile.is_open()){
    std::cout << "Couldn't write features to disk.\n";
    return(-1);
  }
   
  for(size_t i =0; i < fpfhs->points.size(); i++){
    float* hist = fpfhs->points[i].histogram;
    for(size_t j=0; j<33; j++){
      outfile << hist[j];
      if(j<32) outfile << ",";
    }
    outfile << "\n";
  }
  outfile.close();
  
  //Visualize the results
  if(argc > 3){
            
    pcl::visualization::PCLVisualizer viewer("Viewer");
    viewer.setBackgroundColor (0, 0, 0);
    pcl::visualization::PointCloudColorHandlerRGBField<pcl::PointXYZRGB> rgb(cloud);
    viewer.addPointCloud<pcl::PointXYZRGB> (cloud, rgb, "sample cloud");
    viewer.setPointCloudRenderingProperties (pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 3, "sample cloud");
    viewer.addPointCloudNormals<pcl::PointXYZRGB, pcl::Normal> (cloud, normals, 5, 0.005, "normals");
    viewer.addCoordinateSystem (.25);
    viewer.initCameraParameters ();
    while (!viewer.wasStopped ())
    {
      viewer.spinOnce (100);
      boost::this_thread::sleep (boost::posix_time::microseconds (100000));
    }
  }

  return (0);
}
